
#include "DisplayBuffer.h"

volatile uint8_t buf_current_page = 0;
volatile uint8_t buf_current_column = 0;

uint8_t back_buffer[LCD_RAM_PAGES][LCD_WIDTH];
uint8_t front_buffer[LCD_RAM_PAGES][LCD_WIDTH];

uint8_t render_state = ANIM_STOP;

int16_t col_index = 0;


/** Actually, this isn't a classical flip of buffers,
 *  more a synchronization of the front buffer with the back buffer.
 *  The actual work is done by the engine, so we just set the flags.
 */
void flip_buffer(uint8_t animation){
	if(render_state == ANIM_STOP){
	render_state = animation;
		switch(render_state){
			case ANIM_FLIP:
				col_index = 0;
				break;

			case ANIM_PUSH_LEFT:
				col_index = LCD_WIDTH;
				break;

			case ANIM_PUSH_RIGHT:
				col_index = LCD_WIDTH*-1;
				break;
		}
	}
}

//The buffer engine needs to be called frequently
void buffer_engine(void){
	int16_t c,p,cb;
	int16_t shift = 8;
	switch(render_state){
		case ANIM_STOP:
			//Simply do nothing, so return here
			return;
		
		case ANIM_FLIP:
			render_state = ANIM_STOP;
			for(p = 0; p < LCD_RAM_PAGES; p++){
				for(c = 0; c < LCD_WIDTH; c++){
					//Simply copy back to front, no fanciness involved
					front_buffer[p][c] = back_buffer[p][c];
				}
			}
			break;
		
		case ANIM_PUSH_LEFT:
			//Checking for a change of render state
			col_index -= shift;
			if(col_index <= 0){
				render_state = ANIM_FLIP;
				col_index = 0;
			}
			for(p = 0; p < LCD_RAM_PAGES; p++){
				for(c = 0; c < col_index; c++){
					//Simply copy back to front, no fanciness involved
					front_buffer[p][c] = front_buffer[p][c+shift];
				}
			}
			//Insert new columns from right
			for(p = 0; p < LCD_RAM_PAGES; p++){
				cb = 0;
				for(c = col_index; c < LCD_WIDTH; c++){
					//Simply copy back to front, no fanciness involved
					front_buffer[p][c] = back_buffer[p][cb++];
				}
			}
			break;
		
		case ANIM_PUSH_RIGHT:
			//Checking for a change of render state
			col_index += shift;
			if(col_index >= 0){
				render_state = ANIM_FLIP;
				col_index = 0;
			}
			for(p = 0; p < LCD_RAM_PAGES; p++){
				for(c = (LCD_WIDTH-1); c > LCD_WIDTH + col_index; c--){
					//Simply copy back to front, no fanciness involved
					front_buffer[p][c] = front_buffer[p][c-shift];
				}
			}
			//Insert new columns from right
			for(p = 0; p < LCD_RAM_PAGES; p++){
				cb = (col_index *- 1);
				for(c = 0; c < LCD_WIDTH + col_index; c++){
					//Simply copy back to front, no fanciness involved
					front_buffer[p][c] = back_buffer[p][cb++];
				}
			}
			break;
	}
	
	//Writing front buffer to LCD
	for(p = 0; p < LCD_RAM_PAGES; p++){
		lcd_moveto_xy(p, 0);
		for(c = 0; c < LCD_WIDTH; c++){
			lcd_data(front_buffer[p][c]);
		}
	}
}

//Legacy display functions
//change page address by s
uint8_t buf_inc_page(int8_t s) {
  buf_current_page += s;
  return buf_current_page;
}

//change column address by s, 
uint8_t buf_inc_column(int16_t s){
  buf_current_column += s;
  return buf_current_column;
}

//Move cursor
void buf_moveto_xy(uint8_t page, uint8_t column){
	buf_current_column = column; 
	buf_current_page = page;
}

void buf_move_xy(int8_t pages, int16_t columns){
	buf_moveto_xy(buf_inc_page(pages),buf_inc_column(columns));
}


//write data word or command to the LCD
void buf_data(uint8_t data){
	//For memory safety
	if(buf_current_page < LCD_RAM_PAGES && buf_current_column < LCD_WIDTH){
		back_buffer[buf_current_page][buf_current_column] = data;
	}
	buf_inc_column(1);
}

//User functions to write raw data from flash
void buf_put_raw(PGM_P data, uint8_t length){
	buf_inc_column(length);
  while(length--)
    buf_data(pgm_read_byte(data++));
}

/**
 * This function draws a bitmap from the current position on the screen.
 * Parameters:
 * progmem_image - prog_uint8_t array of columns aka the bitmap image
 * pages         - height of image in pages
 * columns       - width of image in pixels (or columns)
 * inverse       - flag which definies if the output is inverted
 */  
void buf_draw_image(const uint8_t *progmem_image, uint8_t pages, uint8_t columns, int8_t inverse) {
	uint16_t i,j;
	uint8_t start_column = buf_current_column;
	//uint16_t x = 0;
	for(j=0;j<pages && buf_current_page < LCD_RAM_PAGES;j++) {
		for (i=0;i<columns && buf_current_column<LCD_WIDTH;i++){
			if(!inverse){
				buf_data(pgm_read_byte_near(&progmem_image[j*columns + i]));
			}else{
				buf_data(~(pgm_read_byte_near(&progmem_image[j*columns + i])));
			}
		}
		if(j+1 != pages){
			buf_moveto_xy(buf_current_page+1,start_column);
		}
	}
}

/**
 * This function draws a bitmap ar any xy-position on the screen.
 * Parameters:
 * progmem_image - prog_uint8_t array of columns aka the bitmap image
 * x             - x start coordinate on the screen
 * y             - y start coordinate on the screen
 * pages         - height of image in pages
 * columns       - width of image in pixels (or columns)
 * inverse       - flag which definies if the output is inverted
 */
void buf_draw_image_xy(const uint8_t *progmem_image, uint8_t x, uint8_t y, uint8_t pages, uint8_t columns, int8_t inverse) {
	uint16_t i,j;
	uint8_t start_column;
	uint8_t offset = y%8;
	uint8_t column;
	//If there is an offset, we must use an additional page
	if(offset){
		pages++;
	}
	buf_moveto_xy((y>>3), x);
	start_column = buf_current_column;
	for(j=0;j<pages && buf_current_page < LCD_RAM_PAGES;j++) {
		for (i=0;i<columns && buf_current_column<LCD_WIDTH;i++){
			column = 0x00;
			if (!offset || j+1 != pages){
				column = (pgm_read_byte_near(&progmem_image[j*columns + i]) << offset);
			}
			if(j > 0){
				column |= ((pgm_read_byte_near(&progmem_image[(j-1)*columns + i])) >> (8-offset));
			}
			if(!inverse){
				buf_data(column);
			}else{
				buf_data(~column);
			}
		}
		if(j+1 != pages){
			buf_moveto_xy(buf_current_page+1,start_column);
		}
	}
}

/**
 * Inverts a given rectangle from a giben starting point and size
 */
void buf_invert(uint8_t start_x, uint8_t start_y, uint8_t width_x, uint8_t height_y){
	uint8_t cur_page, cur_x, page_mask, row, first_page ,last_page;

	//Calc first page
	first_page = start_y / 8;

	//Calc last page
	last_page = (start_y + height_y)/8;
	if(last_page * 8 < start_y + height_y){
		last_page++;
	}

	for(cur_page = first_page;
		cur_page <= last_page && cur_page < LCD_RAM_PAGES;
		cur_page++){

		page_mask = 0xFF;
		//Walking through each pixel in a column of a page
		for(row = 0; row < 8; row++){
			//Cutting the mask writing zeros were no change will be applied
			if(row + (cur_page * 8) < start_y
			|| row + (cur_page * 8) >= start_y + height_y){
				page_mask &= ~(1<<row);
			}
		}
		//Inverting the lines
		for(cur_x = start_x; cur_x < start_x + width_x && cur_x < LCD_WIDTH; cur_x++){
			back_buffer[cur_page][cur_x] = (back_buffer[cur_page][cur_x] & ~page_mask) //None-inverted part
										| ((~(back_buffer[cur_page][cur_x])) & page_mask); //Inverted part
		}
	}
}
