#include "drivers/DS1302.h"

//Defining some useful makros

#define ds_rst_enable()     PORT_DS_RST &= ~_BV(PIN_DS_RST)
#define ds_rst_disable()    PORT_DS_RST |= _BV(PIN_DS_RST)

#define ds_io_input()       DDR_DS_IO &= ~_BV(PIN_DS_IO)
#define ds_io_output()      DDR_DS_IO |= _BV(PIN_DS_IO)

#define ds_io_set_high()    PORT_DS_IO |= _BV(PIN_DS_IO)
#define ds_io_set_low()     PORT_DS_IO &= ~_BV(PIN_DS_IO)

#define ds_clk_high()       PORT_DS_CLK |= _BV(PIN_DS_CLK)
#define ds_clk_low()        PORT_DS_CLK &= ~_BV(PIN_DS_CLK)
#define DS_US_DELAY         2

/** INTERNAL FUNCTIONS **/
uint8_t getDayOfWeek(date indate){
	return (indate.day + (2 * indate.month) + (int) (6 * (indate.month + 1) / 10) + indate.year + (int) (indate.year / 4) - (int) (indate.year / 100) + (int) (indate.year / 400) + 1) % 7; 
}

/**
 * This is the initializer for the DS1302
 */
void ds1302_init(void){
	//Setting the reset pin low - internal clock is still running
	ds_rst_enable();
	//Setting the reset pin as output
	DDR_DS_RST  |= _BV(PIN_DS_RST);
	
	//Configuring clock pin as output
	PORT_DS_CLK |= _BV(PIN_DS_CLK);
	//Clk always starts with low
	ds_clk_low();
	
	//Setting io output to low
	ds_io_set_low();
	ds_io_output();
}

/**
 * This function gets the current date out of the chip
 */
date ds1302_get_date(void){
	uint8_t i,byte_counter,current_byte;     // cc1 and byte_counter are used for counter
	uint8_t burst_data[8];
	date current_date;

	ds_io_output();  // Change IO line to output
	ds_rst_disable(); // Release reset line and prepare for transfer

	current_byte = 0xBF; //0b10111111;    // CLOCK READ BURST COMMAND
	for(i=0;i<8;i++){
		if(current_byte & _BV(i)){
			ds_io_set_high();
		}else{
			ds_io_set_low();
		}
		//Create a clock pulse, about 2us (freq 500Khz)
		ds_clk_low();
		_delay_us(DS_US_DELAY);
		ds_clk_high();
		_delay_us(DS_US_DELAY);
	}
	
	// After sending the command, we need to read all incoming 8 bytes
	ds_io_input();  //Receive data
	
	for(byte_counter=0;byte_counter<8;byte_counter++){
		//current_byte needs to be resetted
		current_byte = 0x00;
		for(i=0;i<8;i++){
			//Create a clock pulse, about 2us (freq 500Khz)
			ds_clk_low();
			_delay_us(DS_US_DELAY);
			
			if(INPUT_DS_IO & (1<<PIN_DS_IO)){
				current_byte |= (1<<i);
			}
			
			ds_clk_high();
			_delay_us(DS_US_DELAY);
		}
		burst_data[byte_counter] = current_byte;  // store result int array
	}
	
	//Set IO line back to default
	ds_clk_low();
	ds_rst_enable(); //Keep on resetting
	ds_io_set_low();
	ds_io_output();  // Change IO line to output
	
	//Parsing the received data - it is stored in a weird BCD format
	current_date.sec = (burst_data[0]&0x0F) + (((burst_data[0]&0x70)>>4)*10);
	current_date.min = (burst_data[1]&0x0F) + (((burst_data[1]&0x70)>>4)*10);
	
	//if(burst_data[2]&0x80){
		//Clock operates in 24h mode - bit 5 serves as +10h indicator
		current_date.hrs = (burst_data[2]&0x0F) + (((burst_data[2]&0x30)>>4)*10);
	/*}else{
		//Clock operates in 12h mode - bit 5 serves as am/pm-indicator
		current_date.hrs = (burst_data[2]&0x1F) + (((burst_data[2]&0x20)>>5)*12);
	}*/
	
	current_date.day = (burst_data[3]&0x0F) + (((burst_data[3]&0x30)>>4)*10);
	current_date.month = (burst_data[4]&0x0F) + (((burst_data[4]&0x10)>>4)*10);
	current_date.dow = (burst_data[5]&0x07);
	current_date.year = (burst_data[6]&0x0F) + (((burst_data[6]&0xF0)>>4)*10) + 2000;
	
	return current_date;
}

/**
 * This function sets the given date to the chip
 */
void ds1302_set_date(date new_date){
	uint8_t i,byte_counter;     // cc1 and byte_counter are used for counter
	uint8_t burst_data[9];
	
	const uint8_t month_length[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	uint8_t length_of_month;
	
	burst_data[0] = 0xBE; //0b10111110;    // WRITE BURST COMMAND
	
	//The DS1302 just accepts about everything, so we need to be sane
	new_date.sec = new_date.sec>200?59:new_date.sec;
	new_date.sec = new_date.sec<60?new_date.sec:0;
	
	new_date.min = new_date.min>200?59:new_date.min;
	new_date.min = new_date.min<60?new_date.min:0;
	
	new_date.hrs = new_date.hrs>200?23:new_date.hrs;
	new_date.hrs = new_date.hrs<24?new_date.hrs:0;
	
	new_date.month = new_date.month>0?new_date.month:12;
	new_date.month = new_date.month<=12?new_date.month:1;
	
	if(new_date.month == 2 && ((new_date.year%4==0 && new_date.year%100!=0)||(new_date.year%400)==0)){
		//Houston, we have a leap year and its february!
		length_of_month = 29;
	}else{
		length_of_month = month_length[(new_date.month-1)];
	}
	new_date.day = new_date.day<=length_of_month?new_date.day:1;
	new_date.day = new_date.day>=1?new_date.day:length_of_month;
	
	//Now convert the RTC-values to the burst data
	burst_data[1] = ((new_date.sec/10)<<4) + (new_date.sec%10);
	burst_data[2] = ((new_date.min/10)<<4) + (new_date.min%10);
	
	burst_data[3] = ((new_date.hrs/10)<<4) + (new_date.hrs%10);
	/*if(new_date.hrs < 12){
		burst_data[3] = (new_date.hrs&0x0F);// | 0x80;
	}else{
		burst_data[3] = ((new_date.hrs-12)&0x0F) | 0x20; //0xA0
	}*/
	burst_data[4] = ((new_date.day/10)<<4) + (new_date.day%10);
	burst_data[5] = ((new_date.month/10)<<4) + (new_date.month%10);
	burst_data[6] = getDayOfWeek(new_date); //Day of Week
	i = (new_date.year-2000);
	burst_data[7] = ((i/10)<<4) + (i%10);
	burst_data[8] = 0x00;
	
	ds_io_output();  // Change IO line to output
	ds_rst_disable(); // Release reset line and prepare for transfer
	
	for(byte_counter=0;byte_counter<9;byte_counter++){
		for(i=0;i<8;i++){
			if(burst_data[byte_counter] & _BV(i)){
				ds_io_set_high();
			}else{
				ds_io_set_low();
			}
			//Create a clock pulse, about 2us (freq 500Khz)
			ds_clk_low();
			_delay_us(DS_US_DELAY);
			ds_clk_high();
			_delay_us(DS_US_DELAY);
		}
	}
	//Set IO line back to default
	ds_clk_low();
	ds_rst_enable(); //Keep on resetting
	ds_io_set_low();
	ds_io_output();  // Change IO line to output
}
