//*****************************************************************************
//
//  File........: RTC.c
//
//  Author(s)...: ATMEL Norway
//
//  Target(s)...: ATmega169
//
//  Compiler....: AVR-GCC 3.3.1; avr-libc 1.0
//
//  Description.: Real Time Clock (RTC)
//
//  Revisions...: 1.0
//
//  YYYYMMDD - VER. - COMMENT                                       - SIGN.
//
//  20021015 - 1.0  - Created                                       - LHM
//  20031009          port to avr-gcc/avr-libc                      - M.Thomas
//  20060602		  Stripped down									- O.Schwaneberg
//  20080219          Major Modifications, added DCF-Support        - O.Schwaneberg
//
//*****************************************************************************

//  Include files
#include "drivers/dcf77.h"

//uint8_t pulse_counter;
date received_time;

//DCFsync und MEZ Information
uint32_t dcfSync = 0;

// Lookup table holding the length of each mont. The first element is a dummy.
// mt this could be placed in progmem too, but the arrays are accessed quite
//    often - so leaving them in RAM is better...
//char MonthLength[13] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

//char LeapMonth;

//Hier wird der dcf String gespeichert
uint8_t dcfData[8] = {0,0,0,0,0,0,0,0};
uint16_t dcfPuls = 0;
uint16_t dcfPause = 0;
uint8_t dcfBit = 0;
uint8_t dcfByte = 0;


/******************************************************************************
*
*   Function name:  RTC_init
*
*   returns:        none
*
*   parameters:     none
*
*   Purpose:        Start Timer/Counter2 in asynchronous operation using a
*                   32.768kHz crystal.
*
*******************************************************************************/
void dcf77_init(void)
{
    //Configuring Hardware
    //Nothing to do. PB4 needs to be input with disabled PullUps
    /*PORTB &= ~_BV(PORTB4);
    DDRB &= ~_BV(PORTB4);*/
}

char parseDcfData(void){
	//VORSICHT: DCF verwendet BIG-ENDIAN!

	//Bit 0 bis 19 sind egal
	received_time.sec = 0;
	received_time.min = 0;
	received_time.hrs = 0;
	received_time.day = 0;
	received_time.month = 0;
	received_time.year = 2000;
	received_time.dow = 0;

	char parity = 0;

	//Bit 17 	„0“: MEZ, „1“: MESZ und Bit 18 	„0“: MESZ, „1“: MEZ
	if(!(dcfData[2] & 0x20) && (dcfData[2] & 0x10)){
		received_time.dst = FALSE;
	}else if((dcfData[2] & 0x20) && !(dcfData[2] & 0x10)){
		received_time.dst = TRUE;
	}else{
		//lcd_putc('Z');
		return FALSE;
	}

	//Bit 19 (dcfData[2] & 0x08) „1“: Am Ende dieser Stunde wird eine Schaltsekunde eingefügt

	//Bit 20 ist immer 1
	if(!(dcfData[2] & 0x04)){
		//sendString("RTC: Warning: Bit 20 = 0!\n");
		//lcd_putc('W');
		return FALSE;
	}

	//Bit 21-27 sind die Minute
	//Zunächst kommen Bit 21-23
	if(dcfData[2]&0x02){
		received_time.min += 1;
		parity++;
	}
	if(dcfData[2]&0x01){
		received_time.min += 2;
		parity++;
	}
	if(dcfData[3]&0x80){
		received_time.min += 4;
		parity++;
	}
	if(dcfData[3]&0x40){
		received_time.min += 8;
		parity++;
	}
	if(dcfData[3]&0x20){
		received_time.min += 10;
		parity++;
	}
	if(dcfData[3]&0x10){
		received_time.min += 20;
		parity++;
	}
	if(dcfData[3]&0x08){
		received_time.min += 40;
		parity++;
	}

	//Bit 28 ist Parität
	if (dcfData[3]&0x04) {	//Prüfbit ok ?
		if (parity == 1 || parity == 3 || parity == 5 || parity == 7) {
			parity = 0;
		}
		else {
			//sendString("DCF: PARITY ERROR: Minute (1)\n");
			//lcd_putc('M');
			return FALSE;
		}
	}else {							//Prüfbit ok ?
		if (parity == 0 || parity == 2 || parity == 4 || parity == 6) {
			parity = 0;
		}
		else{
			//sendString("DCF: PARITY ERROR: Minute (0)\n");
			//lcd_putc('M');
			return FALSE;
		}
	}

	//Bits 29-34 kodieren die Stunde
	if(dcfData[3]&0x02){
		received_time.hrs += 1;
		parity++;
	}
	if(dcfData[3]&0x01){
		received_time.hrs += 2;
		parity++;
	}
	if(dcfData[4]&0x80){
		received_time.hrs += 4;
		parity++;
	}
	if(dcfData[4]&0x40){
		received_time.hrs += 8;
		parity++;
	}
	if(dcfData[4]&0x20){
		received_time.hrs += 10;
		parity++;
	}
	if(dcfData[4]&0x10){
		received_time.hrs += 20;
		parity++;
	}

	//Bit 35 ist Parität
	if (dcfData[4]&0x08) {	//Prüfbit ok ?
		if (parity == 1 || parity == 3 || parity == 5) {
			parity = 0;
		}
		else {
			//sendString("DCF: PARITY ERROR: Hour (1)\n");
			//lcd_putc('H');
			return FALSE;
		}
	}else {							//Prüfbit ok ?
		if (parity == 0 || parity == 2 || parity == 4 || parity == 6) {
			parity = 0;
		}
		else {
			//sendString("DCF: PARITY ERROR: Hour (0)\n");
			//lcd_putc('H');
			return FALSE;
		}
	}

	//Bit 36-41 Monatstag
	if(dcfData[4]&0x04){
		received_time.day += 1;
		parity++;
	}
	if(dcfData[4]&0x02){
		received_time.day += 2;
		parity++;
	}
	if(dcfData[4]&0x01){
		received_time.day += 4;
		parity++;
	}
	if(dcfData[5]&0x80){
		received_time.day += 8;
		parity++;
	}
	if(dcfData[5]&0x40){
		received_time.day += 10;
		parity++;
	}
	if(dcfData[5]&0x20){
		received_time.day += 20;
		parity++;
	}

	//Bit 42-44 Wochentag (0x10 0x08 0x04)
	if(dcfData[5]&0x10){
		received_time.dow += 1;
		parity++;
	}
	if(dcfData[5]&0x08){
		received_time.dow += 2;
		parity++;
	}
	if(dcfData[5]&0x04){
		received_time.dow += 4;
		parity++;
	}

	//Bit 45–49 	5 bit 	Monat (1, 2, 4, 8, 10)
	if(dcfData[5]&0x02){
		received_time.month += 1;
		parity++;
	}
	if(dcfData[5]&0x01){
		received_time.month += 2;
		parity++;
	}
	if(dcfData[6]&0x80){
		received_time.month += 4;
		parity++;
	}
	if(dcfData[6]&0x40){
		received_time.month += 8;
		parity++;
	}
	if(dcfData[6]&0x20){
		received_time.month += 10;
		parity++;
	}

	//Bit 50–57 	8 bit 	Jahr (zweistellig; 1, 2, 4, 8, 10, 20, 40, 80)
	if(dcfData[6]&0x10){
		received_time.year += 1;
		parity++;
	}
	if(dcfData[6]&0x08){
		received_time.year += 2;
		parity++;
	}
	if(dcfData[6]&0x04){
		received_time.year += 4;
		parity++;
	}
	if(dcfData[6]&0x02){
		received_time.year += 8;
		parity++;
	}
	if(dcfData[6]&0x01){
		received_time.year += 10;
		parity++;
	}
	if(dcfData[7]&0x80){
		received_time.year += 20;
		parity++;
	}
	if(dcfData[7]&0x40){
		received_time.year += 40;
		parity++;
	}
	if(dcfData[7]&0x20){
		received_time.year += 80;
		parity++;
	}

	//Bit 58 	1 bit 	Parität Datum (über Bit 36 bis 57)
	if (dcfData[7]&0x10) {	//Prüfbit ok ?
		if (parity%2 == 1) {
			parity = 0;
		}
		else {
			//sendString("DCF: PARITY ERROR: Minute (1)\n");
			//lcd_putc('Y');
			return FALSE;
		}
	}else {							//Prüfbit ok ?
		if (parity%2 == 0) {
			parity = 0;
		}
		else{
			//sendString("DCF: PARITY ERROR: Minute (0)\n");
			//lcd_putc('Y');
			return FALSE;
		}
	}

	/*if(dcfInited){
		uint32_t tempTime = (((uint32_t)received_time.hrs)*3600000) + (((uint32_t)received_time.min)*60000);

		if(getTimeMs() > 85500000){
			dcfSync = 30;
			return;
		}else if(getTimeMs() < 900000){
			dcfSync = 15;
			return;
		}

		//600.000ms = 10min Toleranz
		if(dcfCurMEZ == dcfMEZ){
			if((getTimeMs() > (tempTime+600000)) || (getTimeMs() < (tempTime-600000))){
				return;
			}
		}else{
			if((getTimeMs() > (tempTime+4200000)) || (getTimeMs() < (tempTime-4200000))){
				return;
			}
		}
	}*/
	
	//Die nächste Synchronisation timen in einer Stunde oder 216000 pulses
	dcfSync = 216000;
	
	return TRUE;
}

/**
 * Dieser Handler soll mit 60Hz ausgeführt werden
 */
void dcf77_pulse(void)
{
	//Check if the DCF-Flag is set
	if(eeprom_read_byte((uint8_t*)EE_TIME_SETTINGS) & EE_TIME_FLAG_DCF){
		//Check if the time has come to synchronize
		if(dcfSync == 0){
			if(DCF_PORT & (1<<DCF_PIN)){
				if(dcfPause > 60){
					//Haben wir ein gültiges Signal aufgenommen?
					if(dcfByte == 7 && dcfBit == 3){
						if(parseDcfData()){
							ds1302_set_date(received_time);
						}
					}

					dcfBit = 0;
					dcfByte = 0;
					dcfData[0]=dcfData[1]=dcfData[2]=dcfData[3]=dcfData[4]=dcfData[5]=dcfData[6]=dcfData[7]=0;
				}
				dcfPuls++;
		//			if(dcfPuls > 10){
				dcfPause = 0;
		//			}
			}else{
				if(dcfPause > 20){ //Pause von >333ms
					if(dcfPuls > 2 && dcfPuls < 10){
						//Kurzer Puls (100ms) bedeutet logisch "0"
						dcfBit++;
						if(dcfBit > 7){
							dcfBit = 0;
							dcfByte++;
						}
					}else if(dcfPuls >= 10 && dcfPuls < 16){
						//Langer Puls (200ms) bedeutet logisch "1"
						dcfBit++;
						if(dcfBit > 7){
							dcfBit = 0;
							dcfByte++;
						}
						if(dcfByte < 8){
							dcfData[dcfByte] |= (1<<(7-dcfBit));
						}
					}
					dcfPuls = 0;
				}
				dcfPause++;
			}
		}else if(dcfSync > 0){
			dcfSync--;
		}
	}
}
