#include "drivers/sensors.h"

char ph_inited = FALSE;
float ph_stepsize = 0.0f;
float ph_base_ref = 0.0f;
float ph_base_val = 0.0f;

/**
 * A look up table for the ntc. It contains expected register values for
 * temperatures from -40 to 150 degrees.
 * Note: Usage with Reichelt NTC-0,2 10K.
 **/
const prog_uint16_t temperature_lut[] PROGMEM = {994, 992, 990, 987, 985, 983, 980, 
          977, 975, 972, 969, 965, 961, 958, 954, 950, 946, 941, 937, 932, 
          928, 922, 917, 911, 906, 900, 893, 887, 880, 874, 867, 859, 851, 
          844, 836, 828, 819, 810, 801, 792, 783, 773, 763, 754, 744, 734, 
          723, 713, 702, 692, 681, 670, 659, 648, 637, 626, 615, 603, 592, 
          580, 569, 558, 546, 535, 523, 512, 501, 490, 479, 468, 457, 447, 
          436, 426, 415, 405, 395, 385, 376, 366, 356, 347, 338, 330, 321, 
          312, 304, 296, 287, 279, 271, 264, 257, 250, 243, 236, 230, 223, 
          217, 210, 204, 199, 193, 188, 182, 177, 172, 167, 163, 158, 153, 
          149, 145, 140, 136, 132, 128, 125, 121, 118, 114, 111, 108, 105, 
          102, 99, 96, 94, 91, 89, 86, 84, 82, 79, 77, 75, 73, 71, 69, 67, 
          65, 63, 62, 60, 59, 57, 56, 54, 53, 51, 50, 49, 47, 46, 44, 43, 
          42, 41, 40, 39, 38, 37, 36, 36, 35, 34, 33, 32, 32, 31, 30, 29, 
          28, 28, 27, 26, 25, 25, 24, 24, 23, 23, 22, 22, 21, 21, 20, 20, 
          19, 19, 18};
          
#if (HW_REVISION < 2)
float driver_current_white, driver_current_red;
float driver_current_white_at_startup, driver_current_red_at_startup;
#endif
#if (HW_REVISION >= 2)
float driver_current_white, driver_current_red, driver_current_green, driver_current_blue;
float driver_current_white_at_startup, driver_current_red_at_startup,
      driver_current_green_at_startup, driver_current_blue_at_startup;
#endif
//Supply voltages
#if (HW_REVISION < 2)
float mb_3V3;
#endif
float mb_12V;

//External and system temperature in celsius
int16_t ext_temperature = 0;
int16_t mb_temperature = -300;

//External ph probe
float ext_ph;
float ext_raw_ph;
float samples_ext_ph[PH_OVERSAMPLING];

/*
 *  This Quickselect routine is based on the algorithm described in
 *  "Numerical recipes in C", Second Edition,
 *  Cambridge University Press, 1992, Section 8.5, ISBN 0-521-43108-5
 *  This code by Nicolas Devillard - 1998. Public domain.
 */


#define ELEM_SWAP(a,b) { register float t=(a);(a)=(b);(b)=t; }

float quick_select(float arr[], int n) 
{
	int low, high ;
	int median;
	int middle, ll, hh;

	low = 0 ; high = n-1 ; median = (low + high) / 2;
	for (;;) {
		if (high <= low) /* One element only */
			return arr[median] ;

		if (high == low + 1) {  /* Two elements only */
			if (arr[low] > arr[high])
				ELEM_SWAP(arr[low], arr[high]) ;
			return arr[median] ;
		}

		/* Find median of low, middle and high items; swap into position low */
		middle = (low + high) / 2;
		if (arr[middle] > arr[high])    ELEM_SWAP(arr[middle], arr[high]) ;
		if (arr[low] > arr[high])       ELEM_SWAP(arr[low], arr[high]) ;
		if (arr[middle] > arr[low])     ELEM_SWAP(arr[middle], arr[low]) ;

		/* Swap low item (now in position middle) into position (low+1) */
		ELEM_SWAP(arr[middle], arr[low+1]) ;

		/* Nibble from each end towards middle, swapping items when stuck */
		ll = low + 1;
		hh = high;
		for (;;) {
			do ll++; while (arr[low] > arr[ll]) ;
			do hh--; while (arr[hh]  > arr[low]) ;

			if (hh < ll)
			break;

			ELEM_SWAP(arr[ll], arr[hh]) ;
		}

		/* Swap middle item (in position low) back into correct position */
		ELEM_SWAP(arr[low], arr[hh]) ;

		/* Re-set active partition */
		if (hh <= median)
			low = ll;
		if (hh >= median)
			high = hh - 1;
	}
}

#undef ELEM_SWAP

void readVoltage(char sensor){
	float val = ((float)ADC_read(sensor))/1023.0;
	//val *= 5.0; //For 5.0 volt reference
	
	switch(sensor){
		#if (HW_REVISION < 2)
		case VOLT_3 :
			//Factor 1.47 with 5V Reference
			mb_3V3 = val * 1.47 * 5.0;
			break;
		#endif
		case VOLT_12 :
			//Factor 5.7 with 5V Reference
			mb_12V = val * 5.7 * 5.0;
			break;
	}
}

void readTemperature(char sensor){
	int16_t i;
	uint16_t rval = ADC_read(sensor);
	static uint16_t samples[] = {0,0};
	static uint8_t sample_count[] = {0,0}, last_i[] = {0,0};
	uint8_t j;
	int16_t *temperature;
	#define TEMP_SAMPLES 8
	
	switch(sensor){
		case TEMP_MB:
			j = 0;
			temperature = &mb_temperature;
			break;
		
		case TEMP_EXT:
			j = 1;
			temperature = &ext_temperature;
			break;
		
		default:
			return;
	}
	
	sample_count[j]++;
	samples[j] += rval;
	if(sample_count[j] == TEMP_SAMPLES){ //Enough sample_count collected
		samples[j] /= sample_count[j];
		for (i=0; i<190; i++)  
		{
			if (samples[j] >= pgm_read_word(&temperature_lut[i]))
			{
				/**
				 * Hysteresis:
				 * if i is higher than the last i
				 * and rval - 1 would lead to another value of i
				 * than decrement i by one.
				 */
				if(i > last_i[j] && samples[j] - 1 < pgm_read_word(&temperature_lut[i])){
					i--;
				}
				*temperature = i-40;
				break;
			}
		}
		//saving last i
		last_i[j] = i;
		//Resetting stuff
		samples[j] = 0;
		sample_count[j] = 0;
	}
}

/**
 * Function to measure the current at given led channels
 * Returns TRUE if value has been updated (because of variable update cycles)
 */
char readLEDCurrent(char sensor){
	//Oversampling factor
	#define CURRENT_SAMPLES 31
	//8000 * 68ns = 544µs => at least 14% duty cycle (total cycle length: 4456µs)
	#define CURRENT_MEASURE_DELAY 8000
	//Number of cycles left before deactivation
	#define CURRENT_MEASURE_TLEFT 600
	static float led_current[4][CURRENT_SAMPLES];
	float *p_led_current;
	static uint8_t sample[] = {0, 0, 0, 0};
	float voltage;
	uint8_t index;
	uint16_t counter_limit;
	static volatile uint16_t * timer_counter;
	timer_counter = &TCNT3;
	switch(sensor){
		#if (HW_REVISION >= 2)
		case VOLT_BLUE:
			index = 0;
			counter_limit = BRIGHTNESS_BLUE;
			p_led_current = &driver_current_blue;
			break;
		case VOLT_GREEN:
			index = 1;
			counter_limit = BRIGHTNESS_GREEN;
			p_led_current = &driver_current_green;
			break;
		#endif
		case VOLT_RED:
			index = 2;
			counter_limit = BRIGHTNESS_RED;
			p_led_current = &driver_current_red;
			break;
		default: //white
			index = 3;
			sensor = VOLT_WHITE;
			counter_limit = BRIGHTNESS_WHITE;
			p_led_current = &driver_current_white;
			timer_counter = &TCNT1;
			break;
	}
	/*
	 * The LEDs must be active during a current measurement.
	 * We do not try to measure at the very beginning or the end of a pulse.
	 * We need at least 3% duty-cycle to perform measurements.
	 */
	if(counter_limit > CURRENT_MEASURE_DELAY + CURRENT_MEASURE_TLEFT 
	 && *timer_counter < (counter_limit-CURRENT_MEASURE_TLEFT)
	 && *timer_counter > CURRENT_MEASURE_DELAY){
		//Getting the voltage at desired LED-FET-Driver
		ADMUX = sensor;
		ADCSRA |= (1<<ADSC);        // do single conversion
		_delay_us(50);
		//Check if timer_counter is still below limit
		if(*timer_counter < counter_limit){
			while(!(ADCSRA & 0x10));    // wait for conversion done, ADIF flag active
			voltage = (float)ADC/204.6f; //Directly convert to voltage
			if(sample[index] < CURRENT_SAMPLES){
				if(sensor == VOLT_WHITE){
					led_current[index][sample[index]] = (voltage / ( RDS_IRF520_RS * (1.0f + (RDS_IRF520_DRIFT * ((float)mb_temperature - 30.0f)))));
				}else{
					led_current[index][sample[index]] = (voltage / RDS_IRF520_RS);
				}
				sample[index]++;
			}
			if(sample[index] == CURRENT_SAMPLES){
				*p_led_current = quick_select(led_current[index],CURRENT_SAMPLES); //led_current[index] / (float)sample[index];
				if(*p_led_current < 0.01){
					*p_led_current = 0.01; //Avoid 0
				}
				sample[index] = 0;
				return TRUE;
			}
		}
	}
	return FALSE;
}

/**
 * Primitive version without further checks
 */
void readInitialLEDCurrent(char sensor){
	uint8_t i;
	#define INIT_SAMPLES 25
	float voltage, current[INIT_SAMPLES], result;
	float *p_led_current, *p_initial_led_current;
	
	switch(sensor){
		#if (HW_REVISION >= 2)
		case VOLT_BLUE:
			p_led_current = &driver_current_blue;
			p_initial_led_current = &driver_current_blue_at_startup;
			break;
		case VOLT_GREEN:
			p_led_current = &driver_current_green;
			p_initial_led_current = &driver_current_green_at_startup;
			break;
		#endif
		case VOLT_RED:
			p_led_current = &driver_current_red;
			p_initial_led_current = &driver_current_red_at_startup;
			break;
		default: //white
			sensor = VOLT_WHITE;
			p_led_current = &driver_current_white;
			p_initial_led_current = &driver_current_white_at_startup;
			break;
	}
	
	for (i = 0; i < INIT_SAMPLES; i++){
		
		voltage = (float)ADC_read(sensor) / 204.6f; //Directly convert to voltage
		if(sensor == VOLT_WHITE){
			current[i] = (voltage / ( RDS_IRF520_RS * (1.0f + (RDS_IRF520_DRIFT * ((float)mb_temperature - 30.0f)))));
		}else{
			current[i] = (voltage / RDS_IRF520_RS);
		}
		_delay_ms(1);
	}
	
	//Finding median
	result = quick_select(current, INIT_SAMPLES);
	
	if(result < 0.01){
		result = 0.01; //Avoid 0
	}
	
	*p_led_current = *p_initial_led_current = result;
	
	*p_initial_led_current *= 1.2f; //Add 20% tolerance
	*p_initial_led_current += 0.2f; //Add a constant offset
}

void initPH(void){
	float ph_ref1, ph_ref2;
	int16_t ph_val1, ph_val2;
	ph_ref1 = ((float) eeprom_read_word((uint16_t*)EE_PH_REF1)) / 100.0f;
	ph_ref2 = ((float) eeprom_read_word((uint16_t*)EE_PH_REF2)) / 100.0f;
	ph_val1 = eeprom_read_word((uint16_t*)EE_PH_VAL1);
	ph_val2 = eeprom_read_word((uint16_t*)EE_PH_VAL2);
	
	//PH change per digit / LSB
	ph_stepsize = (ph_ref2 - ph_ref1) / (float)(ph_val2 - ph_val1);
	
	//Setting base values
	ph_base_ref = ph_ref1;
	ph_base_val = ph_val1;
	
	ph_inited = TRUE;
}

void readPH(char sensor){
	static uint16_t oversample_counter = 0;
	//static uint32_t samples_ext_ph = 0;
	//int i;
	//Values need to be set
	if(!ph_inited){
		initPH();
	}
	
	if(oversample_counter < PH_OVERSAMPLING){
		//Oversampling
		//for(i = 0; i < PH_SAMPLES_PER_CALL; i++){
		samples_ext_ph[oversample_counter++] = (float) ADC_read(sensor);
		//oversample_counter++;
		//}
	}else{
		ext_raw_ph = quick_select(samples_ext_ph, PH_OVERSAMPLING);
		float val = ext_raw_ph;
		//Reset oversampling stuff
		oversample_counter = 0;
		
		//float val = (float)ext_ph_val;
		val = ((val - ph_base_val) * ph_stepsize) + ph_base_ref;
		
		ext_ph = val;
	}
}
