//=============================================================================
//Display Library
//for EA-DOGM132 GLCD (132px x 32px)
//and EA-DOGM128 GLCD (128px x 64px)
//
//Provides all basic functions to access the display
//Since no graphics ram is used, the memory footprint is rather small but also
//does not allow to change single pixels. Data written to the LCD can not be
//read back!
//Text can be displayed using the attached font generator containing several
//character sets and font variants.
//
//Author:  Jan Michel (jan at mueschelsoft dot de)
//License: GNU General Public License, version 3
//Version: v0.9 July 2009
//=============================================================================


#include "dogm-graphic.h"

//=============================================================================
//keeping track of current position in ram - necessary for big fonts & bitmaps
//=============================================================================

volatile uint8_t lcd_current_page = 0;
volatile uint8_t lcd_current_column = 0;


//change page address by s
uint8_t lcd_inc_page(int8_t s) {
  lcd_current_page += s;
  while (lcd_current_page >= LCD_RAM_PAGES) 
    lcd_current_page -= LCD_RAM_PAGES;
  return lcd_current_page;
  }

//change column address by s, 
uint8_t lcd_inc_column(int16_t s) {
  lcd_current_column += s;
//  while (lcd_current_column >= LCD_WIDTH)
//    lcd_current_column -= LCD_WIDTH;
  return lcd_current_column;
  }


//Move cursor to new position and update internal address
void lcd_moveto_xy(uint8_t page, uint8_t column) {
	LCD_GOTO_ADDRESS(page,column);
	lcd_current_column = column; 
	lcd_current_page = page;
}

void lcd_move_xy(int8_t pages, int16_t columns) {
	lcd_moveto_xy(lcd_inc_page(pages),lcd_inc_column(columns));
}


//=============================================================================
//Basic Byte Access to Display
//=============================================================================

//Writes one byte to the data ram
void lcd_data(uint8_t data) {
  /*spi_write(0xFF);
  spi_wait_for_idle();*/
  LCD_SELECT();
  LCD_DRAM();
  spi_write(data);
  spi_wait_for_idle();
  LCD_UNSELECT();
  lcd_inc_column(1);
  }

//Writes one command byte
void lcd_command(uint8_t cmd) {
  /*spi_write(0xFF);
  spi_wait_for_idle();*/
  LCD_SELECT();
  LCD_CMD();
  spi_write(cmd);
  spi_wait_for_idle();
  LCD_UNSELECT();
  }
  

//=============================================================================
//Puts raw data from Flash to the Display
//=============================================================================
void lcd_put_raw (PGM_P data, uint8_t length) {
  lcd_inc_column(length);
  while(length--)
    lcd_data(pgm_read_byte(data++));
  }

/**
 * This function draws a bitmap from the current position on the screen.
 * Parameters:
 * progmem_image - prog_uint8_t array of columns aka the bitmap image
 * pages         - height of image in pages
 * columns       - width of image in pixels (or columns)
 * inverse       - flag which definies if the output is inverted
 */  
void lcd_draw_image(const uint8_t *progmem_image, uint8_t pages, uint8_t columns, int8_t inverse) {
	uint16_t i,j;
	uint8_t start_column = lcd_current_column;
	//uint16_t x = 0;
	for(j=0;j<pages && lcd_current_page < LCD_RAM_PAGES;j++) {
		for (i=0;i<columns && lcd_current_column<LCD_WIDTH;i++){
			if(!inverse){
				lcd_data(pgm_read_byte_near(&progmem_image[j*columns + i]));
			}else{
				lcd_data(~(pgm_read_byte_near(&progmem_image[j*columns + i])));
			}
		}
		if(j+1 != pages){
			lcd_moveto_xy(lcd_current_page+1,start_column);
		}
	}
}

/**
 * This function draws a bitmap ar any xy-position on the screen.
 * Parameters:
 * progmem_image - prog_uint8_t array of columns aka the bitmap image
 * x             - x start coordinate on the screen
 * y             - y start coordinate on the screen
 * pages         - height of image in pages
 * columns       - width of image in pixels (or columns)
 * inverse       - flag which definies if the output is inverted
 */
void lcd_draw_image_xy(const uint8_t *progmem_image, uint8_t x, uint8_t y, uint8_t pages, uint8_t columns, int8_t inverse) {
	uint16_t i,j;
	uint8_t start_column;
	uint8_t offset = y%8;
	//If there is an offset, we must use an additional page
	if(offset){
		pages++;
	}
	lcd_moveto_xy((y>>3), x);
	start_column = lcd_current_column;
	for(j=0;j<pages && lcd_current_page < LCD_RAM_PAGES;j++) {
		for (i=0;i<columns && lcd_current_column<LCD_WIDTH;i++){
			uint8_t column = (pgm_read_byte_near(&progmem_image[j*columns + i]) << offset);
			if(j > 0){
				column |= ((pgm_read_byte_near(&progmem_image[(j-1)*columns + i])) >> (8-offset));
			}
			if(!inverse){
				lcd_data(column);
			}else{
				lcd_data(~column);
			}
		}
		if(j+1 != pages){
			lcd_moveto_xy(lcd_current_page+1,start_column);
		}
	}
}

//=============================================================================
//initializes the display in 6 o'clock mode, 4x booster for 2.4-3.3V supply voltage
//scheme according to datasheet
//Compatible to both Display types: 128x64 and 132x32
//=============================================================================
void lcd_init() {
  DDR_A0  |= _BV(PIN_A0);             //set outputs
  DDR_RST |= _BV(PIN_RST);
  LCD_UNSELECT();
  //PORTC |= _BV(PC3); //We need to take care of CS-SD to keep the sd-card quiet
  //DDR_CS  |= _BV(PIN_CS) | _BV(PC3); //by Oliver S.
  
  /* Set MOSI, SCK and ^SS output, all others input */
  /*DDRB = (1<<PB2)|(1<<PB1)|(1<<PB0);
  //init spi: msb first, update falling, read rising, clk/2=7.3728 MHz
  SPCR = (0 << SPIE) | (1 << SPE) | (0 << DORD) | (1 << MSTR) | (0 << CPOL) | (0 << CPHA) | (0 << SPR1) | (0 << SPR0);
  SPSR = (1 << SPI2X);
  SPDR = 0;
  
  spi_dummywrite(); //Get rid of junk in the pipe*/
  
  LCD_RESET_ON();                     //Apply Reset to the Display Controller
  _delay_ms(1);
  LCD_RESET_OFF();

  LCD_SET_FIRST_LINE(0);              //first bit in RAM is on the first line of the LCD
  LCD_SET_BOTTOM_VIEW();              //6 o'clock mode, normal orientation
  LCD_ORIENTATION_NORMAL();
  LCD_SET_MODE_POSITIVE();            //positive display
  //LCD_SET_MODE_INVERTED();
  LCD_SET_BIAS_RATIO_1_9();           //bias 1/9
  //LCD_SET_POWER_CONTROL(7);           //power control mode: all features on
  LCD_SET_LOW_POWER();                //Activate low power mode
  LCD_SET_BIAS_VOLTAGE(3);            //set voltage regulator R/R
  LCD_SET_VOLUME_MODE(0x1F);          //volume mode set
  LCD_SET_INDICATOR_OFF();            //switch indicator off, no blinking
  LCD_SWITCH_ON();                    //Switch display on
  uint8_t i,j;                        //clear display content
  for(j=0;j<8;j++) {
    lcd_moveto_xy(j,0);
    for (i=0;i<132;i++)
      lcd_data(0);
    }
  lcd_moveto_xy(0,0);                 //Set write pointer
  LCD_SET_FIRST_LINE(0);              //first bit in RAM is on the first line of the LCD


  return;
  }
